#!/usr/bin/ruby -w
# -*- ruby -*-

require 'log' unless Log::VERSION


# A thing that can be grepped.
class InputFile
  include Loggable

  attr_reader :lines, :fname, :stati, :count

  # cross-platform end of line:   DOS  UNIX  MAC
  ANY_END_OF_LINE = Regexp.new(/(?:\r\n|\n|\r)/)
  
  def initialize(fname, lines)
    @fname = fname
    @lines = lines
    @stati = Array.new          # index = line number, value = context character
    if $options && $options.count
      @count = 0
    else
      @count = nil
    end
    @output = $options && $options.grep_output ? GrepOutputFormat.new(self) : GlarkOutputFormat.new(self)
    @extracted = $/ == "\n" ? @lines : nil
    @regions   = nil
  end

  def has_context
    $options.after != 0 || $options.before != 0
  end

  def set_status(from, to, ch, force = false)
    from.upto(to) do |ln|
      if force or not @stati[ln]
        @stati[ln] = ch
      end
    end
  end

  def mark_as_match(start_line, end_line = start_line)
    if $options.grep_output
      end_line = start_line
    end

    $options.exit_status = $options.invert_match ? 1 : 0

    if @count
      @count += 1
    else
      st = [0, start_line - $options.before].max
      set_status(st,           start_line - 1,            "-")
      set_status(start_line,   end_line,                  ":",  true)
      set_status(end_line + 1, end_line + $options.after, "+")
    end
  end

  def write_matches
    @output.write_matches
  end

  def write_all
    @output.write_all
  end

  def write_non_matches
    @output.write_non_matches
  end

  def calculate_regions
  end

  # Returns the lines for this file, separated by end of line sequences.
  def get_lines
    unless @extracted
      @extracted = []
      
      # This is much easier. Just resplit the whole thing at end of line sequences
      
      eoline = "\n"             # should be OS-dependent
      reallines = @lines.join("").split(ANY_END_OF_LINE)
      
      # "\n" after all but the last line
      (0 ... (reallines.length - 1)).each do |lnum|
        @extracted << reallines[lnum] + eoline
      end
      @extracted << reallines[-1]
    end
    @extracted
  end

  # Returns the given line for this file. For this method, a line ends with a
  # CR, as opposed to the "lines" method, which ends with $/.
  def get_line(lnum)
    log "lnum: #{lnum}"
    get_lines()[lnum]
  end

  # returns the range that is represented by the region number
  def get_range(rnum)
    unless @regions
      if $/ == "\n"
        @regions = (0 ... @lines.length).collect { |lnum| lnum .. lnum }
      else
        @regions = []           # keys = region number; values = range of lines

        lstart = 0
        @lines.each do |line|
          lend = lstart
          log "considering <<#{line.gsub(/\n/, '\\n')}>>"
          log "line.chomped: <<#{line.chomp.gsub(/\n/, '\\n')}>>"
          line.scan(ANY_END_OF_LINE).each do |cr|
            log "cr: #{cr}"
            lend += 1
          end

#          log "adding region #{lstart} .. #{lend}"
          @regions << Range.new(lstart, lend - 1)

          lstart = lend
        end
        
#         @regions.each_with_index do |range, idx|
#           log "region[#{idx}] = #{range}"
#         end
      end
    end

    @regions[rnum]
  end
end


class OutputFormat
  include Loggable

  def initialize(infile)
    @infile = infile
    @show_file_name = ARGV.size > 0 && (ARGV.size > 1 || FileTester.type(ARGV[0]) == FileTester::DIRECTORY) && $options.show_file_names
  end
  
  # prints the line, and adjusts for the fact that in our world, lines are
  # 0-indexed, whereas they are displayed as if 1-indexed.
  def print_line(lnum, ch = nil)
    log "lnum #{lnum}, ch: '#{ch}'"
    begin
      lnums = @infile.get_range(lnum)
      log "lnums(#{lnum}): #{lnums}"
      if lnums
        lnums.each do |ln|
          if $options.show_line_numbers
            printf "%5d ", ln + 1
          end
          if ch && has_context
            printf "%s ", ch
          end
          puts @infile.get_line(ln)
        end
      end
    rescue => e
      puts e
      puts e.backtrace
    end
  end

  def has_context
    $options.after != 0 || $options.before != 0
  end

  def write_matches
    if $options.count
      write_count
    else
      lastln = nil
      0.upto(@infile.stati.size) do |ln|
        if @infile.stati[ln]
          print_line(ln, @infile.stati[ln]) 
          lastln = ln
        end
      end
    end
  end

  def write_non_matches
    if $options.count
      write_non_count
    else
      (0 ... @infile.lines.length).each do |ln|
        unless @infile.stati[ln] && @infile.stati[ln] == ":"
          print_line(ln) 
        end
      end
    end
  end

  def write_all
    (0 ... @infile.lines.length).each do |ln|
      print_line(ln) 
    end
  end

end


class GlarkOutputFormat < OutputFormat

  def show_file_header
    print $options.file_highlight if $options.highlight
    print @infile.fname, ":"
    print ANSIColor.reset if $options.highlight
    print "\n"
  end

  def write_count
    puts "    " + @infile.count.to_s
  end

  def write_non_count
    puts "    " + (@infile.lines.size - @infile.count).to_s
  end

  def write_matches
    show_file_header if @show_file_name
    super
  end

  def write_non_matches
    show_file_header if @show_file_name
    super
  end

  def write_all
    show_file_header if @show_file_name
    super
  end

end


# This matches grep, mostly. It is for running within emacs, thus,
# it does not support context or highlighting.
class GrepOutputFormat < OutputFormat

  def write_count
    print @infile.fname, ":" if @show_file_name
    puts @infile.count
  end

  def write_non_count
    print @infile.fname, ":" if @show_file_name
    puts @infile.lines.length - @infile.count
  end

  # prints the line, and adjusts for the fact that in our world, lines are
  # 0-indexed, whereas they are displayed as if 1-indexed.
  def print_line(lnum, ch = nil)
    print @infile.fname, ":" if @show_file_name
    if $options.show_line_numbers
      printf "%d: ", lnum + 1
    end
    print @infile.lines[lnum]
  end

end


class BinaryFile < InputFile

  def write_matches
    if $options.count
      write_count
    else
      puts "Binary file " + @fname + " matches"
    end
  end

  def write_non_matches
    if $options.count
      write_non_count
    else
      puts "Binary file " + @fname + " matches"
    end
  end

end


if __FILE__ == $0
  AppLog.verbose = true
  require 'glarkfile'
  require 'glarkopt'
  go = GlarkOptions.new
  go.count = true

  f = InputFile.new("glarkformat", [ "foo\n", "bar\n" ])
  f.mark_as_match(0, 14)
  puts "matches:"
  f.write_matches
  puts "end of matches."

  go.grep_output = true

  f = InputFile.new("grepformat", %w{ foo bar baz })
  f.mark_as_match(0, 14)
  puts "matches:"
  f.write_matches
  puts "end of matches."

end
